import logging

from argparse import ArgumentParser
from pathlib import Path

import feedparser
import toml

from .downloader import downloaders


def main():
    """Downloads RSS updates based on command line arguments."""

    # Parse args
    parser = ArgumentParser(description="Download RSS updates")

    parser.add_argument('-n', '--num-entries', type=int, default=0,
        metavar='N', help="Download last N updates")
    parser.add_argument('-s', '--subfolders', action='store_true',
        help="Store each feed's updates in its own subfolder")
    parser.add_argument('-d', '--downloader', type=str, default='meta_refresh',
        help="Download handler")
    parser.add_argument('-m', '--mkpath', action='store_true',
        help="Create destination path, if it doesn't exist")
    parser.add_argument('-v', '--verbose', action='count', default=0,
        help="Print messages to console")
    parser.add_argument('-r', '--redownload', action='store_true',
        help="Ignore cache and redownload all entries")
    parser.add_argument('srcs', nargs='+', metavar='src',
        help="URL of an RSS or Atom feed")
    parser.add_argument('dest', type=Path,
        help="Directory feeds are downloaded to")

    args = parser.parse_args()

    # Setup logger
    logging.basicConfig(
        level = logging.WARNING - 10 * args.verbose,
        format = "%(message)s")

    # Mkpath option
    if args.mkpath:
        args.dest.mkdir(parents=True, exist_ok=True)

    # Open/create cache
    cache_path = args.dest.joinpath('.cache').with_suffix('.toml')

    if cache_path.exists() and not args.redownload:
        cache = toml.load(cache_path)
    else:
        cache = {}

    # Set user-agent string
    feedparser.USER_AGENT = "RSSsync/0.1"

    # Download feeds
    for src in args.srcs:
        cache.setdefault(src, {})
        downloader = downloaders[args.downloader](src, args.dest, cache[src])

        if args.verbose:
            logging.info(f"Updating {src}...")

        downloader.download_feed()

        if downloader.is_feed_modified:
            if args.subfolders:
                downloader.create_subfolder()

            downloader.download_entries(args.num_entries)
            logging.info("Done.")
        else:
            logging.info("Feed is up to date.")

    # Save cache
    with cache_path.open('w') as f:
        toml.dump(cache, f)



"""
TODO:
- Downloaders from .config/rsssync/config.py
- Add metadata to entries via xattrs
- Allow renaming subfolder
- Get RSS feed URLs from YouTube channels, etc.
- Only download entries that match query
- Stay open and update feeds periodically
- Man page
- Sphinx docs
- List downloaders in help text
- Option to simply print out feed metadata
- Set default downloader based on URL in config.py
- Import/export OPML
- Start date option
- Notifications for new updates
- Human-readable cache
- Make cache depend on files
- Multiple downloaders based on eg whether an update has an enclosure
- Handle redirects
"""


from .downloader import FeedDownloader, downloaders
from . import meta_refresh_downloader, pls_downloader



from .downloader import FeedDownloader, downloaders


class PlsDownloader(FeedDownloader):
    """Downloads podcast episodes as individual PLS files"""

    def download_entry(self, entry):
        if 'enclosures' in entry and len(entry.enclosures) > 0:
            filename = FeedDownloader.fix_filename(entry.title)
            path = self.dest.joinpath(filename + '.pls')

            with path.open('w') as f:
                contents = \
                    '[playlist]\n' \
                    'File1={}\n' \
                    'NumberOfEntries=1'

                f.write(contents.format(entry.enclosures[0].href))

            self.attach_metadata(path, entry)


downloaders['pls'] = PlsDownloader



from .downloader import FeedDownloader, downloaders


class MetaRefreshDownloader(FeedDownloader):
    """Downloads updates as HTML files that redirect to the update's link"""

    def download_entry(self, entry):
        filename = FeedDownloader.fix_filename(entry.title)
        path = self.dest.joinpath(filename + '.html')

        with path.open('w') as f:
            contents = \
                '<html><head>' \
                '<meta http-equiv="refresh" content="0; url={}"/>' \
                '</head></html>'

            f.write(contents.format(entry.link))

        self.attach_metadata(path, entry)


downloaders['meta_refresh'] = MetaRefreshDownloader



import logging
import re

from os import utime
from pathlib import Path
from time import mktime
from urllib.error import URLError

import feedparser


class FeedDownloader():
    """Handles the downloading of a feed.

    Attributes:
        src (str): URL of the feed.
        dest (Path): Path where feed will be downloaded.
        feed (FeedParserDict): The feed, once it's downloaded.
        cache (dict): Info to prevent redownloading feeds/entries.
    """

    def fix_filename(filename):
        """Removes forbidden characters from filename"""

        return re.sub(r'[\\/*?:"<>|]', '', filename)


    def __init__(self, src, dest, cache):
        self.src = src
        self.dest = dest
        self.feed = {}
        self.cache = cache


    @property
    def new_entries(self):
        return [e for e in self.feed.entries if not self.is_cached(e)]


    @property
    def has_response(self):
        return 'status' in self.feed


    @property
    def has_good_response(self):
        return self.feed.get('status') == 200 or self.feed.get('status') == 304


    @property
    def is_feed_modified(self):
        return self.feed.get('status') == 200 and len(self.new_entries) > 0


    def is_cached(self, entry):
        return entry.get('guid') in self.cache.get('entries', {}) \
            or entry.get('title') in self.cache.get('entries', {})


    def download_feed(self):
        """Downloads the RSS/Atom feed itself."""

        try:
            self.feed = feedparser.parse(
                self.src,
                etag=self.cache.get('etag'),
                modified=self.cache.get('modified'))

            if self.has_good_response:
                if 'etag' in self.feed:
                    self.cache['etag'] = self.feed.etag
                if 'modified' in self.feed:
                    self.cache['modified'] = self.feed.modified

            elif self.has_response:
                status = self.feed.status
                logging.error(f"Error {status}: Feed couldn't be downloaded.")

            else:
                logging.error("Error: No response.")

        except URLError:
            logging.error("Error: URL could not be resolved.")


    def create_subfolder(self):
        """Creates subfolder for feed entries."""

        folder_name = FeedDownloader.fix_filename(self.feed.feed.title)
        self.dest = self.dest.joinpath(folder_name)
        self.dest.mkdir(exist_ok=True)


    def download_entries(self, num_entries):
        """Iterates over entries in the feed, calling download_entry on each.

        Probably won't need to be overridden, unless you want to download
        entries in parallel.

        Args:
            num_entries (int): Number of entries to download.
        """
 
        entries = self.feed.entries[:num_entries] \
                if num_entries > 0 else self.feed.entries
        self.cache.setdefault('entries', [])

        for e in self.new_entries:
            self.download_entry(e)
            self.cache['entries'].append(e.get('guid') or e.get('title'))


    def attach_metadata(self, path, entry):
        """Attaches various metadata to the file representing an entry."""

        # Set date modified/accessed to publication date
        pub_time = mktime(entry.published_parsed)
        utime(path, (pub_time, pub_time))


    def download_entry(self, entry):
        """Downloads an individual entry.

        This method must be overridden to actually create any files. Note that
        in many cases, it's perfectly acceptable to simply create a file
        containing a link to the actual entry -- no downloading required.

        Args:
            entry (FeedParserDict): Data on the entry to download.
        """

        raise NotImplementedError()


downloaders = {}
